/*
 *  Copyright 2001 Adrian Thurston <adriant@ragel.ca>
 */

/*  This file is part of Ragel.
 *
 *  Ragel is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  Ragel is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with Ragel; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

#include <string.h>
#include <assert.h>
#include "fsmgraph.h"

/* Insert a transition into an inlist. The head must be supplied. Also sets
 * fromState and toState. */
void FsmAp::attachTrans( StateAp *from, StateAp *to, 
		TransAp *&head, TransAp *trans )
{
	/* Assert that the state is not attached. */
	assert( trans->fromState == 0 );
	assert( trans->toState == 0 );

	trans->next = head;
	trans->prev = 0;

	/* If in trans list is not empty, set the head->prev to trans. */
	if ( head != 0 )
		head->prev = trans;

	/* Now insert ourselves at the front of the list. */
	head = trans;

	/* The transition is now attached. Remember the parties involved. */
	trans->fromState = from;
	trans->toState = to;

	/* Keep track of foreign transitions for from and to. */
	if ( from != to ) {
		if ( misfitAccounting ) {
			/* If the number of foreign in transitions is about to go up to 1 then
			 * move it from the misfit list to the main list. */
			if ( to->foreignInTrans == 0 )
				stateList.append( misfitList.detach( to ) );
		}
		
		to->foreignInTrans += 1;
	}
};

/* Detach a transition from an inlist. The head of the inlist must be supplied.
 * Also reset the transitions's fromState and toState pointers. */
void FsmAp::detachTrans( StateAp *from, StateAp *to, 
		TransAp *&head, TransAp *trans )
{
	/* Assert that the state is indeed attached. */
	assert( trans->fromState == from );
	assert( trans->toState == to );

	/* Detach in the inTransList. */
	if ( trans->prev == 0 ) 
		head = trans->next; 
	else
		trans->prev->next = trans->next; 

	if ( trans->next != 0 )
		trans->next->prev = trans->prev; 
	
	/* A detached state does not link any states. */
	trans->fromState = 0;
	trans->toState = 0;

	/* Keep track of foreign transitions for from and to. */
	if ( from != to ) {
		to->foreignInTrans -= 1;
		
		if ( misfitAccounting ) {
			/* If the number of foreign in transitions goes down to 0 then move it
			 * from the main list to the misfit list. */
			if ( to->foreignInTrans == 0 )
				misfitList.append( stateList.detach( to ) );
		}
	}
}

/* Attach states on the default transition, range list or on out/in list key.
 * Type of attaching and the onChar used is controlled by keyType. First makes
 * a new transition. If there is already a transition out from fromState on the
 * default, then will assertion fail. */
TransAp *FsmAp::attachNewTrans( StateAp *from, StateAp *to, 
		FsmKeyType keyType, const long &lowKey, const long &highKey )
{
	/* Make the new transition. */
	TransAp *retVal = addTrans();

	switch ( keyType ) {
	case KeyTypeSingle: {
		assert( false );
		break;
	}
	case KeyTypeRange: {
		/* Make the entry in the out list for the transitions. */
		TransEl *outTel = from->outRange.insert( lowKey, highKey, keyOps );
		assert( outTel != 0 );

		/* Set the out list pointer to the new trans. */
		outTel->value = retVal;

		/* Set the lower key in the trans so we can map the trans back to the
		 * record in the outRange. */
		retVal->lowerKey = lowKey;

		/* Attach using inRange as the head pointer. */
		attachTrans( from, to, to->inRange.head, retVal );
		break;
	}
	case KeyTypeDefault: {
		/* If there is already a default, then assertion fail. */
		assert( from->outDefault == 0 );

		/* Attach using the new transition and defInTrans as the head. */
		from->outDefault = retVal;

		/* The lower key is not used for default transitions, but init it for
		 * consistency anyways. */
		retVal->lowerKey = 0;

		/* Attach using inDef as the head pointer. */
		attachTrans( from, to, to->inDefault.head, retVal );
		break;
	}}

	return retVal;
}

/* Attach for out/in lists, range lists or for default transition. Type of
 * attaching is controlled by the keyType parameter. This attach should be
 * used when a transition already is allocated and it must be attached. */
void FsmAp::attachExisting( StateAp *from, StateAp *to, TransAp *trans, 
		FsmKeyType keyType, const long &onChar )
{
	switch ( keyType ) {
	case KeyTypeSingle: {
		assert( false );
		break;
	}
	case KeyTypeRange: {
		/* Set the lower key in the trans so we can map the trans back to the
		 * record in the outRange. */
		trans->lowerKey = onChar;

		/* Attach using the inRange pointer as the head pointer. */
		attachTrans( from, to, to->inRange.head, trans );
		break;
	}
	case KeyTypeDefault: {
		/* Lower char is not used. */
		/* trans->lowerKey = NOTHING; */

		/* Attach using inDef as the head pointer. */
		attachTrans( from, to, to->inDefault.head, trans );
		break;
	}}
}

/* Detach for out/in lists or for default transition. The type of detaching is
 * controlled by the keyType parameter. */
void FsmAp::detachStates( StateAp *from, StateAp *to, TransAp *trans, 
		FsmKeyType keyType, const long &onChar )
{
	switch ( keyType ) {
	case KeyTypeSingle: {
		assert( false );
		break;
	}
	case KeyTypeRange: {
		/* Detach using to's inRange pointer as the head. */
		detachTrans( from, to, to->inRange.head, trans );
		break;
	}
	case KeyTypeDefault: {
		/* Detach using the to's inDefHead as the head pointer. */
		detachTrans( from, to, to->inDefault.head, trans );
		break;
	}}
}


/* Detach a state from the graph. Detaches and deletes transitions in and out
 * of the state. Empties inList and outList. Removes the state from the final
 * state set. A detached state becomes useless and should be deleted. */
void FsmAp::detachState( StateAp *state )
{
	/* Detach the in transitions from the inRange list of transitions. */
	while ( state->inRange.head != 0 ) {
		/* Get pointers to the trans and the state. */
		TransAp *trans = state->inRange.head;
		StateAp *fromState = trans->fromState;

		/* Detach the transitions from the source state. */
		detachStates( fromState, state, trans, KeyTypeRange, trans->lowerKey );

		/* Find the transition element in from's out range. It should
		 * always be there. . */
		TransEl *fromTel = fromState->outRange.find( trans->lowerKey, keyOps );
		assert( fromTel != 0 );

		/* If from has a default transition we cannot delete the key-value
		 * pair from it's out range because that effectively adds the keys to
		 * the set of keys following the default trans.  Instead null out
		 * the transition pointer. */
		if ( fromState->outDefault != 0 ) {
			/* From has a default transition, we must null out the trans ptr. */
			fromTel->value = 0;
		}
		else {
			/* Remove two items at pos fromTel. Uses Vector::remove. */
			int pos = fromTel - fromState->outRange.data;
			fromState->outRange.TransListVect::remove( pos );
		}

		/* Finished with the transition. */
		deleteTrans( trans );
		delete trans;
	}

	/* Detach default in transitions. */
	while ( state->inDefault.head != 0 ) {
		/* Get pointers to the trans and the state. */
		TransAp *trans = state->inDefault.head;
		StateAp *fromState = trans->fromState;

		/* Detach. */
		detachStates( fromState, state, trans, KeyTypeDefault, 0 );

		/* From state no longer has a default trans, null it out. */
		fromState->outDefault = 0;

		/* Finsished with the trans. */
		deleteTrans( trans );
		delete trans;
	}

	/* Remove the entry points in on the machine. */
	while ( state->entryIds.length() > 0 )
		unsetEntry( state->entryIds[0], state );

	/* Detach out range transitions. */
	for ( TransList::Iter tel = state->outRange; tel.lte(); tel++ ) {
		if ( tel->value != 0 ) {
			detachStates( state, tel->value->toState, tel->value,
					KeyTypeRange, tel->lowKey );
			deleteTrans( tel->value );
			delete tel->value;
		}
	}

	/* Delete all of the out range pointers. */
	state->outRange.empty();

	/* Detach the default out transition. */
	if ( state->outDefault != 0 ) {
		detachStates( state, state->outDefault->toState, state->outDefault, 
				KeyTypeDefault, 0 );
		deleteTrans( state->outDefault );
		delete state->outDefault;
	}

	/* Unset final stateness before detaching from graph. */
	if ( state->stateBits & SB_ISFINAL )
		finStateSet.remove( state );
}


/* Duplicate a transition. Makes a new transition that is attached to the same
 * dest as srcTrans. The new transition has functions and priority taken from
 * srcTrans. If leavingFsm is set, then outPrior and outFuncs are taken from
 * the from state. Used for merging a transition in to a free spot. The trans
 * can just be dropped in. It does not conflict with an existing trans and need
 * not be crossed. Returns the new transition. */
TransAp *FsmAp::dupTrans( StateAp *from, FsmKeyType keyType, const long &onChar,
			TransAp *srcTrans, bool leavingFsm )
{
	/* Make a new transition. */
	TransAp *newTrans = addTrans();

	/* We can attach the transition, one does not exist. */
	attachExisting( from, srcTrans->toState, newTrans, keyType, onChar );
		
	/* Call the user callback to add in the original source transition. */
	addInTrans( newTrans, srcTrans );

	/* Add in the from state's out data. This will pickup any out functions or
	 * and/or set the priority from the out prior. This must happen after the
	 * adding in of the trans, so that out priority overrides the priority
	 * from the source trans. */
	if ( leavingFsm )
		leavingFromState( newTrans, from );

	return newTrans;
}


/* In crossing, src trans overwrites the existing one because it has a
 * higher priority. The existing destTrans is deleted and replaced with
 * a copy of srcTrans. */
TransAp *FsmAp::overwriteTrans( MergeData &md, StateAp *from, FsmKeyType keyType, 
			const long &onChar, TransAp *destTrans, TransAp *srcTrans, 
			bool leavingFsm )
{
	/* Toast the existing transition. */
	detachStates( from, destTrans->toState, destTrans, keyType, onChar );

	/* Callback to indicate the transition is about the be delted. */
	deleteTrans( destTrans );

	/* Delete the existing transition, it is being replaced and make a brand
	 * new one to replace it. This is for completeness of concept. Fsm graphs
	 * built on top of this template expect it. */
	delete destTrans;
	TransAp *newTrans = addTrans();

	/* We can attach the transition, one does not exist. */
	attachExisting( from, srcTrans->toState, newTrans, keyType, onChar );

	/* Add in src trans to the new transition. We don't need to merge in the
	 * from out trans data, that was done already. */
	addInTrans( newTrans, srcTrans );

	/* Return the replacement transition. */
	return newTrans;
}

/* In crossing, src trans and dest trans go to two different states. Make one
 * state from the sets of states that src and dest trans go to. */
TransAp *FsmAp::fsmAttachStates( MergeData &md, StateAp *from, FsmKeyType keyType, 
			const long &onChar, TransAp *destTrans, TransAp *srcTrans, 
			bool leavingFsm )
{
	/* The priorities are equal. We must merge the transitions. Does the
	 * existing trans go to the state we are to attach to? ie, are we to
	 * simply double up the transition? */
	StateAp *toState = srcTrans->toState;
	StateAp *existingState = destTrans->toState;

	if ( existingState == toState ) {
		/* The transition is a double up to the same state.  Copy the src
		 * trans into itself. We don't need to merge in the from out trans
		 * data, that was done already. */
		addInTrans( destTrans, srcTrans );
	}
	else {
		/* The trans is not a double up. Dest trans cannot be the same as src
		 * trans. Set up the state set. */
		StateSet stateSet;

		/* We go to all the states the existing trans goes to, plus... */
		if ( existingState->stateDictEl == 0 )
			stateSet.insert( existingState );
		else
			stateSet.insert( existingState->stateDictEl->stateSet );

		/* ... all the states that we have been told to go to. */
		if ( toState->stateDictEl == 0 )
			stateSet.insert( toState );
		else
			stateSet.insert( toState->stateDictEl->stateSet );

		/* Look for the state. If it is not there already, make it. */
		StateDictEl *lastFound;
		if ( md.stateDict.insert( stateSet, &lastFound ) ) {
			/* Make a new state representing the combination of states in
			 * stateSet. It gets added to the fill list.  This means that we
			 * need to fill in it's transitions sometime in the future.  We
			 * don't do that now (ie, do not recurse). */
			StateAp *combinState = addState();

			/* Link up the dict element and the state. */
			lastFound->targState = combinState;
			combinState->stateDictEl = lastFound;

			/* Add to the fill list. */
			md.fillListAppend( combinState );
		}

		/* Get the state insertted/deleted. */
		StateAp *targ = lastFound->targState;

		/* Detach the state from existing state. */
		detachStates( from, existingState, destTrans, keyType, onChar );

		/* Re-attach to the new target. */
		attachExisting( from, targ, destTrans, keyType, onChar );

		/* Add in src trans to the existing transition that we redirected to
		 * the new state. We don't need to merge in the from out trans data,
		 * that was done already. */
		addInTrans( destTrans, srcTrans );
	}

	return destTrans;
}

/* Find the trans with the higher priority. If src is lower priority then dest then
 * src is ignored. If src is higher priority than dest, then src overwrites dest. If
 * the priorities are equal, then they are merged. */
TransAp *FsmAp::crossTransitions( MergeData &md, StateAp *from, FsmKeyType keyType, 
		const long &onChar, TransAp *destTrans, TransAp *srcTrans, 
		bool leavingFsm )
{
	TransAp *retTrans, *dupSrcTrans;

	/* Set up a dummy transition that will serve as the source transition.
	 * This duplicate will be set as an out trans from the from state and then
	 * it will be used for the cross. This lets the source pick up any out
	 * transition data of from before priorities are compared. */
	dupSrcTrans = addTrans();
	dupSrcTrans->fromState = srcTrans->fromState;
	dupSrcTrans->toState = srcTrans->toState;

	/* Add in the original source transition. */
	addInTrans( dupSrcTrans, srcTrans );

	/* Add in the from state's out data. This will pickup any out functions or
	 * and/or set the priority from the out prior. It is important that this
	 * happens after the adding in of the trans, so that out priority
	 * overrides the priority from the source trans. */
	if ( leavingFsm )
		leavingFromState( dupSrcTrans, from );

	/* Compare the priority of the dest and src transitions. */
	int compareRes = comparePrior( destTrans, dupSrcTrans );
	if ( compareRes < 0 ) {
		/* Src trans has a higher priority than dest, src overwrites dest. */
		retTrans = overwriteTrans( md, from, keyType, onChar, destTrans, 
				dupSrcTrans, leavingFsm );
	}
	else if ( compareRes > 0 ) {
		/* The existing trans has a higher priority. We do nothing. No new transition
		 * is made. Return destTrans. */
		retTrans = destTrans;
	}
	else {
		/* Src trans and dest trans have the same priority, they must be merged. */
		retTrans = fsmAttachStates( md, from, keyType, onChar, destTrans, 
				dupSrcTrans, leavingFsm );
	}

	/* Finished with this temporary. */
	deleteTrans( dupSrcTrans );
	delete dupSrcTrans;

	/* Return the transition that resulted from the cross. */
	return retTrans;
}

/* When copying out transitions, a transition exists only in the dest out list.
 * Since the key is not in src, we must consider a default transition.  Dest
 * may get crossed with it. */
TransAp *FsmAp::keyInDestEl( MergeData &md, StateAp *dest, StateAp *src, 
		TransEl *destTel, FsmKeyType keyType, TransAp *defTrans, bool leavingFsm )
{
	TransAp *retTrans = 0;

	/* Is there a default transition? */
	if ( defTrans != 0 ) {
		/* There is a default, is destTel set? */
		if ( destTel->value != 0 ) {
			/* Cross dest trans with the default from src. */
			retTrans = crossTransitions( md, dest, keyType, destTel->lowKey, 
					destTel->value, defTrans, leavingFsm );
		}
		else {
			/* Dest trans el is not set, copy the default from src in. */
			retTrans = dupTrans( dest, keyType, destTel->lowKey, defTrans, leavingFsm );
		}
	}
	else {
		/* No default transition from src. Just use what was in dest already. */
		retTrans = destTel->value;		
	}
	return retTrans;
}

/* When copying out transitions, a transition exists only in the src out list.
 * Since the key is not in dest, we must consider a default transition. Src may
 * get crossed with one of those. */
TransAp *FsmAp::keyInSrcEl( MergeData &md, StateAp *dest, StateAp *src, 
		TransEl *srcTel, FsmKeyType keyType, TransAp *defTrans, bool leavingFsm )
{
	TransAp *retTrans = 0;

	if ( defTrans != 0 ) {
		/* Dest has default trans. Src's trans may be crossed with it. */
		if ( srcTel->value != 0 ) {
			/* Cross src trans with the default. First get a copy of the default
			 * to pretend as if it was already in dest on the key. */
			TransAp *newTrans = dupTrans( dest, keyType, srcTel->lowKey,
					defTrans, false );
			retTrans = crossTransitions( md, dest, keyType, srcTel->lowKey,
					newTrans, srcTel->value, leavingFsm );
		}
		else {
			/* Dup dest's default into retTrans. */
			retTrans = dupTrans( dest, keyType, srcTel->lowKey, defTrans, false );
		}
	}
	else {
		/* Dest has no default transition. */
		if ( srcTel->value != 0 ) {
			/* Dup src's trans into the new transition element. */
			retTrans = dupTrans( dest, keyType, srcTel->lowKey, 
					srcTel->value, leavingFsm );
		}
		else {
			/* Src is explicitly off. New transition element must also be off
			 * explicitly. */
			retTrans = 0;
		}
	}
	return retTrans;
}

/* When copying out transitions, a transition exists (on the same key) in both
 * src and dest. We need to determine if the transitions are explicitly unset.
 * No crossing with any default transition or with range transitions will
 * happen. Note that the default or range are assumed only when there is no
 * key. */
TransAp *FsmAp::keyInBothEl( MergeData &md, StateAp *dest, 
		StateAp *src, TransEl *destTel, TransEl *srcTel, FsmKeyType keyType, 
		bool leavingFsm )
{
	TransAp *retTrans = 0;

	if ( destTel->value != 0 && srcTel->value != 0 ) {
		/* Transitions are set in both, must cross them. */
		retTrans = crossTransitions( md, dest, keyType, destTel->lowKey,
				destTel->value, srcTel->value, leavingFsm );
	}
	else if ( destTel->value != 0 ) {
		/* TransAp is set only in destTel. Continue using destTel. */
		retTrans = destTel->value;
	}
	else if ( srcTel->value != 0 ) {
		/* TransAp is set only in srcTel. Dup it. */
		retTrans = dupTrans( dest, keyType, srcTel->lowKey, srcTel->value, leavingFsm );
	}
	else {
		/* TransAp is unset in both. Stays unset.  */
		retTrans = 0;
	}

	return retTrans;
}


/* Copying the default transition. */
void FsmAp::copyDefTrans( MergeData &md, StateAp *dest, StateAp *src, bool leavingFsm )
{
	if ( dest->outDefault != 0 && src->outDefault != 0 ) {
		/* Must cross the default out transitions. */
		dest->outDefault = crossTransitions( md, dest, KeyTypeDefault, 0,
					dest->outDefault, src->outDefault, leavingFsm );
	}
	else if ( src->outDefault != 0 ) {
		/* A default trans only in src, copy it to dest. */
		dest->outDefault = dupTrans( dest, KeyTypeDefault, 0,
				src->outDefault, leavingFsm );
	}
	else {
		/* Leave dest->outDefault as is. */
	}
}

void FsmAp::changeRangeLowerKey( TransAp *trans, const long &oldKey, 
		const long &newKey )
{
	/* Be certain that the the we are changing from the correct old key and
	 * set the new key. */
	assert( keyOps->eq( trans->lowerKey, oldKey ) );
	trans->lowerKey = newKey;
}

/* Copy the transitions in srcList to the outlist of dest. If srcList comes from
 * a final state and out transitions should be copied in then leavingFsm can be
 * set true. SrcList should not be the outList of dest, otherwise you would
 * be copying the contents of srcList into itself as it's iterated: bad news. */
void FsmAp::outTransCopy( MergeData &md, StateAp *dest, StateAp *src, bool leavingFsm )
{
	/* For making destination lists. */
	TransEl newTel;

	/* The destination list. */
	TransList destRange;

	/* Set up an iterator to stop at breaks. */
	PairIter outPair(dest, src, keyOps, true);
	while ( ! outPair.end() ) {
		switch ( outPair.userState ) {
		case PairIter::RangeInS1:
			/* Ranges may get crossed with src's default transition. */
			newTel.value = keyInDestEl( md, dest, src, outPair.s1Tel,
					KeyTypeRange, outPair.defTrans, leavingFsm );

			/* Add the new range to the new out range. */
			newTel.lowKey = outPair.s1Tel->lowKey;
			newTel.highKey = outPair.s1Tel->highKey;
			destRange.append( newTel );
			break;

		case PairIter::RangeInS2:
			/* Ranges may get crossed with dests's default transition. */
			newTel.value = keyInSrcEl( md, dest, src, outPair.s2Tel, 
					KeyTypeRange, outPair.defTrans, leavingFsm );

			/* Add newTel to the new out range. Get the keys from srcTel, easy. */
			newTel.lowKey = outPair.s2Tel->lowKey;
			newTel.highKey = outPair.s2Tel->highKey;
			destRange.append( newTel );
			break;

		case PairIter::RangeOverlap:
			/* Call worker for transitions in both using range type. */
			newTel.value = keyInBothEl( md, dest, src, outPair.s1Tel, 
					outPair.s2Tel, KeyTypeRange, leavingFsm );

			/* Add newTel to the new out range. Get the keys from destTel. Keys could
			 * come from either dest or src. */
			newTel.lowKey = outPair.s1Tel->lowKey;
			newTel.highKey = outPair.s1Tel->highKey;
			destRange.append( newTel );
			break;

		case PairIter::BreakS1:
			/* The source range may be null, in which case, no need to copy. */
			if ( outPair.s1Tel->value != 0 ) {
				/* The dest trans needs to be copied. The copy goes into the
				 * first half of the break because that is the only value we
				 * have access to. */
				outPair.s1Tel->value = dupTrans( dest, KeyTypeRange, 
						outPair.s1Tel->lowKey, outPair.s1Tel->value, false );

				/* The lower char is the key of in transitions and the next
				 * trans is having its lower char changed. This must be
				 * reflected by the move of the transition to the new in
				 * transition list. */
				changeRangeLowerKey( outPair.s1NextTel.value, 
						outPair.s1Tel->lowKey, outPair.s1NextTel.lowKey );
			}
			break;

		case PairIter::BreakS2:
			/* We do not care about duplicating transitions from src. They are 
			 * merely sourced so do not need to be duplicated to protected 
			 * against writing twice. */
			break;
		}
		outPair++;
	}

	/* Last thing to do is copy default transitions. */
	copyDefTrans( md, dest, src, leavingFsm );

	/* Transfer destList into dest->outList. */
	dest->outRange.empty();
	dest->outRange.shallowCopy( destRange );
	destRange.abandon();
}


/* Move all the transitions that go into src so that they go into dest.  */
void FsmAp::inTransMove( StateAp *dest, StateAp *src )
{
	/* Do not try to move in trans to and from the same state. */
	assert( dest != src );

	/* If src is the start state, dest becomes the start state. */
	if ( src == startState ) {
		unsetStartState();
		setStartState( dest );
	}

	/* For each entry point into, create an entry point into dest, when the
	 * state is detached, the entry points to src will be removed. */
	for ( EntryIdSet::Iter enId = src->entryIds; enId.lte(); enId++ )
		changeEntry( *enId, dest, src );

	/* Move the transitions in inRange. */
	while ( src->inRange.head != 0 ) {
		/* Get trans and from state. */
		TransAp *trans = src->inRange.head;
		StateAp *fromState = trans->fromState;

		/* Detach from src, reattach to dest. */
		detachStates( fromState, src, trans, KeyTypeRange, trans->lowerKey );
		attachExisting( fromState, dest, trans, KeyTypeRange, trans->lowerKey );
	}

	/* Move the default transitions. */
	while ( src->inDefault.head != 0 ) {
		/* Get trans and from state. */
		TransAp *trans = src->inDefault.head;
		StateAp *fromState = trans->fromState;

		/* Detach from src, reattach to dest. */
		detachStates( fromState, src, trans, KeyTypeDefault, 0 );
		attachExisting( fromState, dest, trans, KeyTypeDefault, 0 );
	}
}
