/*
 *  Copyright 2001-2004 Adrian Thurston <adriant@ragel.ca>
 */

/*  This file is part of Ragel.
 *
 *  Ragel is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  Ragel is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with Ragel; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

%{

#include <iostream>
#include <stdlib.h>
#include <limits.h>
#include <errno.h>
#include "ragel.h"
#include "parsetree.h"
#include "buffer.h"

using std::endl;

/* Buffer that a code block is collected into. */
Buffer codeBlockBuf;

/* The parse data. For each fsm spec, the parser collects things that it parses
 * in data structures in here. */
ParseData *pd;

/* Bison 1.875 tries to prevent a warning, but results ends up producing a 
 * This is an ugly hack. But it works for this case. */
#ifdef BISON_V1875
#define __attribute__(arg)
#endif

/* Try to do a definition, common to assignment and instantiation. */
void tryMachineDef( const BISON_YYLTYPE &loc, char *name, 
		Join *join, bool isInstance );

%}

%union {
	/* General data types. */
	char c;
	char *data;
	int integer;
	Literal *literal;

	/* Tree nodes. */
	Term *term;
	FactorWithAug *factorWithAug;
	FactorWithRep *factorWithRep;
	FactorWithNeg *factorWithNeg;
	Factor *factor;
	Expression *expression;
	Join *join;

	/* Priorities and actions. */
	AugType augType;
	Action *action;
	PriorDesc *priorDesc;

	/* Regular expression items. */
	RegExpr *regExp;
	ReItem *reItem;
	ReOrBlock *reOrBlock;
	ReOrItem *reOrItem;
}

%token TK_Section

/* General tokens. */
%token <data> TK_UInt
%token <data> TK_PInt
%token <data> TK_NInt
%token <data> TK_Hex
%token <data> TK_Word
%token <data> TK_Literal
%token TK_DoubleDot
%token TK_Arrow
%token TK_ColonEquals
%token TK_StartError
%token TK_AllError
%token TK_FinishError
%token TK_LeaveError

/* Regular expression tokens. */
%token RE_Slash
%token RE_SqOpen
%token RE_SqOpenNeg
%token RE_SqClose
%token RE_Dot
%token RE_Star
%token RE_Dash
%token <data> RE_Char

/* Tokens specific to inline code. */
%token <data> IL_WhiteSpace
%token <data> IL_Comment
%token <data> IL_Literal
%token <data> IL_Symbol

/* Keywords. */
%token KW_Data
%token KW_Action
%token KW_Init
%token KW_AlphType
%token KW_Range

/* Specials in code blocks. */
%token KW_Hold
%token KW_Char
%token KW_Goto
%token KW_Call
%token KW_Ret
%token KW_Stack

/* Symbols. In ragel lexical space, the scanner does not pass 
 * any data along with the symbols, in inline code lexical 
 * space it does. */
%token '*' '?' '+' '!' '(' ')' ';' ',' '=' 
%token ':' '@' '%' '$' '-' '|' '&' '.' '>'

%type <augType> AugTypeBase
%type <augType> AugTypeError
%type <integer> PriorityAug
%type <data> PriorityAugNum
%type <action> ActionEmbed
%type <join> Join
%type <expression> Expression
%type <term> Term
%type <factorWithAug> FactorWithAug
%type <factorWithAug> FactorWithLabel
%type <factorWithRep> FactorWithRep
%type <integer> FactorRepNum
%type <factorWithNeg> FactorWithNeg
%type <factor> Factor
%type <literal> RangeLit
%type <data> AlphabetNum
%type <data> MachineName
%type <integer> PriorityName

/* Curly blocks are not interpreted. Code blocks and struct blocks are. */
%type <data> CurlyBlock
%type <data> CurlyBlockAny
%type <data> CurlyBlockSpecial
%type <data> CodeBlock
%type <data> CodeBlockAny
%type <data> CodeBlockSpecial
%type <data> StructBlock
%type <data> StructBlockAny
%type <data> StructBlockSymbol

%type <regExp> RegularExpr
%type <reItem> RegularExprItem
%type <reItem> RegularExprChar
%type <reOrBlock> RegularExprOrData
%type <reOrItem> RegularExprOrChar

%%

/* Input is any number of input sections. An empty file is accepted. */
input: FsmSpecList;
FsmSpecList: 		
	FsmSpecList FsmSpec { } |
	{ };

/* Fsm Specification. Fsms are deliminated by the %% tokens 
 * and require a name for the fsm immediately following the open %% 
 * deliminator. */
FsmSpec:
	TK_Section TK_Word StatementList TK_Section {
		/* Set the fsm's name. */
		pd->fsmName = $2;
		pd->fsmStartSecLoc = InputLoc(@1);
		pd->fsmEndSecLoc = InputLoc(@4);

		/* If there were no errors, then generate output. */
		if ( pd->errorCount == 0 ) {
			/* Generate output, dot file or fsm spec. */
			switch ( outputFormat ) {
			case OutCCode: case OutCppCode: 
				pd->generateCode( );
				break;
			case OutGraphvizDot:
				pd->generateGraphviz( );
				break;
			case OutPrint:
				pd->printFsm();
				break;
			case OutDump:
				pd->dumpFsm();
				break;
			}
		}
		delete pd;

		/* Make a parse data for the next fsm machine. */
		pd = new ParseData( inputFile );
	};

/* Garble up a to the end of a section. */
FsmSpec: TK_Section error TK_Section { yyerrok; };


/* A NonEmpty list of statements in a fsm. */
StatementList:
	StatementList Statement |
	/* Nothing */;

/* The differnt types of statements in a fsm spec. */
Statement:	
	Assignment |
	Instantiation |
	DataSpec |
	ActionSpec |
	InitSpec |
	AlphSpec |
	RangeSpec;

/* Garble up to the next ; */
Statement: error ';' { yyerrok; };

/* Allow the user to give some code to insert into the init function. */
InitSpec:
	KW_Init CurlyBlock {
		// Put the code on the init code list. 
		pd->initCodeList.append( new StringListEl( InputLoc(@2), $2 ) );
	} ;

/* Allow the user to give some declarations to put into the fsm structure This
 * block is interpreted. */
DataSpec:	
	KW_Data StructBlock ';' {
		// Put the spec for the data on the list. 
		pd->dataList.append( new StringListEl( 
				InputLoc(@2), $2 ) );
	} ;

/* Allow the user to create a named fsm action that can be referenced when
 * building a machine. */
ActionSpec:	
	KW_Action TK_Word CodeBlock {
		if ( pd->actionDict.find( $2 ) ) {
			/* Recover by just ignoring the duplicate. */
			error(@2) << "action \"" << $2 << "\" already defined" << endl;
		}
		else {
			/* Add the action to the list of actions. */
			Action *newAction = new Action( InputLoc(@3), $2, $3, 
					pd->nextActionId++ );

			/* Insert to list and dict. */
			pd->actionList.append( newAction );
			pd->actionDict.insert( newAction );
		}
	};

/* Specifies the data type of the input alphabet. One or two words 
 * followed by a semi-colon. */
AlphSpec:
	KW_AlphType TK_Word TK_Word ';' {
		if ( ! pd->setAlphType( $2, $3 ) ) {
			// Recover by ignoring the alphtype statement.
			error(@2) << "\"" << $2 << 
				" " << $3 << "\" is not an allowed alphabet type" << endl;
		}
	} |
	KW_AlphType TK_Word ';' {
		if ( ! pd->setAlphType( $2 ) ) {
			// Recover by ignoring the alphtype statement.
			error(@2) << "\"" << $2 << "\" is not an allowed alphabet type" << endl;
		}
	};

/* Specifies a range to assume that the input characters will fall into. */
RangeSpec:
	KW_Range AlphabetNum AlphabetNum ';' {
		// Save the upper and lower ends of the range and emit the line number.
		pd->lowerNum = $2;
		pd->upperNum = $3;
		pd->rangeLowLoc = InputLoc(@2);
		pd->rangeHighLoc = InputLoc(@3);
	};

/* An assignement statement. Assigns the definition of a machine to a variable name. */
Assignment:
	MachineName '=' Join ';' {
		bool isInstance = false;

		/* Main machine must be an instance. */
		if ( strcmp($1, machineMain) == 0 ) {
			warning(@1) << "main machine will be implicitly instantiated" << endl;
			isInstance = true;
		}

		/* Generic creation of machine for instantiation and assignment. */
		tryMachineDef( @1, $1, $3, isInstance );
	};

/* An instantiation statement. Instantiates a machine and assigns it to a
 * variable name. */
Instantiation:
	MachineName TK_ColonEquals Join ';' {
		/* Generic creation of machine for instantiation and assignment. */
		tryMachineDef( @1, $1, $3, true );
	};

/* Capture the machine name for making the machine's priority name. */
MachineName:
	TK_Word {
		// Make the machine available as a priority name. The name may be used already.
		PriorDictEl *priorDictEl;
		if ( pd->priorDict.insert( $1, pd->nextPriorKey, &priorDictEl ) )
			pd->nextPriorKey += 1;

		// Set the prior num inserted/found as the current default priority
		// number.
		pd->curDefPriorKey = priorDictEl->value;
	};

Join: 
	Join ',' Expression {
		/* Append the expression to the list and return it. */
		$1->exprList.append( $3 );
		$$ = $1;
	} |
	Expression {
		/* Create the expression list with the intial expression. */
		$$ = new Join( @1, $1 );
	};

/* Top level production in the parse of a fsm. The lowest precedence
 * is the '|' (or), '&' (intersection), and '-' (subtraction) operators. */
Expression:
	Expression '|' Term {
		$$ = new Expression( $1, $3, Expression::OrType );
	} |
	Expression '&' Term {
		$$ = new Expression( $1, $3, Expression::IntersectType );
	} |
	Expression '-' Term {
		$$ = new Expression( $1, $3, Expression::SubtractType );
	} |
	Term {
		$$ = new Expression( $1 );
	};

Term:
	Term FactorWithAug {
		$$ = new Term( $1, $2 );
	} |
	Term '.' FactorWithAug {
		$$ = new Term( $1, $3 );
	} |
	FactorWithAug {
		$$ = new Term( $1 );
	};

/* Third group up in precedence. Allow users to embed actions and to change the
 * priorities of transitions. */
FactorWithAug:
	FactorWithAug TK_Arrow TK_Word { 
		/* Add the target to the list and return the factor object. */
		$1->epsilonLinks.append( EpsilonLink( @3, $3 ) );
		$$ = $1; 
	} |
	FactorWithAug AugTypeBase ActionEmbed {
		/* Append the action to the factorWithAug, record the refernce from 
		 * factorWithAug to the action and pass up the factorWithAug. */
		$1->actions.append( ParserAction( $2, $3 ) );
		$3->actionRefs.append( $1 );
		$$ = $1;
	} |
	FactorWithAug AugTypeError ActionEmbed {
		/* Append the action to the factorWithAug, record the refernce from 
		 * factorWithAug to the action and pass up the factorWithAug. */
		$1->actions.append( ParserAction( $2, $3 ) );
		$3->actionRefs.append( $1 );
		$$ = $1;
	} |
	FactorWithAug AugTypeBase '(' PriorityName ',' PriorityAug ')' {
		// Append the priority using a default name.
		$1->priorityAugs.append( PriorityAug( $2, $4, $6 ) );
		$$ = $1;
	} |
	FactorWithAug AugTypeBase PriorityAug {
		// Append the named priority to the factorWthAug and pass it up.
		$1->priorityAugs.append( PriorityAug( $2, pd->curDefPriorKey, $3 ) );
		$$ = $1;
	} |
	FactorWithLabel {
		$$ = $1;
	};

/* A specified priority name. Looks up the name in the current priority
 * dictionary. */
PriorityName:
	TK_Word {
		// Lookup/create the priority key.
		PriorDictEl *priorDictEl;
		if ( pd->priorDict.insert( $1, pd->nextPriorKey, &priorDictEl ) )
			pd->nextPriorKey += 1;

		// Use the inserted/found priority key.
		$$ = priorDictEl->value;
	};

/* Priority change specs. */
PriorityAug: 
	PriorityAugNum {
		// Convert the priority number to a long. Check for overflow.
		errno = 0;
		int aug = strtol( $1, 0, 10 );
		if ( errno == ERANGE && aug == LONG_MAX ) {
			// Priority number too large. Recover by setting the priority to 0.
			error(@1) << "priority number " << $1 << " overflows" << endl;
			$$ = 0;
		}
		else if ( errno == ERANGE && aug == LONG_MIN ) {
			// Priority number too large in the neg. Recover by using 0.
			error(@1) << "priority number " << $1 << " underflows" << endl;
			$$ = 0;
		}
		else {
			// No overflow or underflow.
			$$ = aug;
 		}
	};

PriorityAugNum:
	TK_NInt | TK_PInt | TK_UInt ;


/* Differnt places to change priorities. */
AugTypeBase:
	'@' { $$ = at_fin; } |
	'%' { $$ = at_leave; } |
	'$' { $$ = at_all; } |
	'>' { $$ = at_start; };
		
/* Differnt places to embed actions. */
AugTypeError:
	TK_StartError { $$ = at_start_error; } |
	TK_AllError { $$ = at_all_error; } |
	TK_FinishError { $$ = at_fin_error; } |
	TK_LeaveError { $$ = at_leave_error; };


/* Different ways to embed actions. A TK_Word is reference to an action given by
 * the user as a statement in the fsm specification. An action can also be
 * specified immediately. */
ActionEmbed:
	TK_Word {
		/* Set the name in the actionDict. */
		Action *action = pd->actionDict.find( $1 );
		if ( action != 0 ) {
			/* Pass up the action element */
			$$ = action;
		}
		else {
			// Recover by returning null as the action.
			error(@1) << "action lookup of \"" << $1 << "\" failed" << endl;
			$$ = 0;
		}
	} |
	CodeBlock {
		/* Create the action, add it to the list and pass up. */
		Action *newAction = new Action( InputLoc(@1), 0, $1, pd->nextActionId++ );
		pd->actionList.append( newAction );
		$$ = newAction;
	};


/* Allow the labelling of machines. Included in the third 
 * level of precedence. */
FactorWithLabel: 
	TK_Word ':' FactorWithLabel { 
		/* Add the label to the list and pass the factor up. */
		$3->labels.prepend( $1 );
		$$ = $3; 
	} |
	FactorWithRep {
		$$ = new FactorWithAug( $1 );
	}

/* The fourth level of precedence. These are the trailing unary operators that
 * allow for repetition. */
FactorWithRep:
	FactorWithRep '*' {
		$$ = new FactorWithRep( InputLoc(@2), $1, 0, 0,
				FactorWithRep::StarType );
	} |
	FactorWithRep '?' {
		$$ = new FactorWithRep( InputLoc(@2), $1, 0, 0,
				FactorWithRep::OptionalType );
	} |
	FactorWithRep '+' {
		$$ = new FactorWithRep( InputLoc(@2), $1, 0, 0,
				FactorWithRep::PlusType );
	} |
	FactorWithRep '{' FactorRepNum '}' {
		$$ = new FactorWithRep( InputLoc(@2), $1, $3, 0,
				FactorWithRep::ExactType );
	} |
	FactorWithRep '{' ',' FactorRepNum '}' {
		$$ = new FactorWithRep( InputLoc(@2), $1, 0, $4,
				FactorWithRep::MaxType );
	} |
	FactorWithRep '{' FactorRepNum ',' '}' {
		$$ = new FactorWithRep( InputLoc(@2), $1, $3, 0,
				FactorWithRep::MinType );
	} |
	FactorWithRep '{' FactorRepNum ',' FactorRepNum '}' {
		$$ = new FactorWithRep( InputLoc(@2), $1, $3, $5,
				FactorWithRep::RangeType );
	} |
	FactorWithNeg {
		$$ = new FactorWithRep( InputLoc(@1), $1 );
	};

FactorRepNum:
	TK_UInt {
		// Convert the priority number to a long. Check for overflow.
		errno = 0;
		int rep = strtol( $1, 0, 10 );
		if ( errno == ERANGE && rep == LONG_MAX ) {
			// Repetition too large. Recover by returing repetition 1. */
			error(@1) << "repetition number " << $1 << " overflows" << endl;
			$$ = 1;
		}
		else {
			// Cannot be negative, so no overflow.
			$$ = rep;
 		}
	};

/* The fifth level up in precedence. Negation. */
FactorWithNeg:
	'!' FactorWithNeg {
		$$ = new FactorWithNeg( InputLoc(@1), $2 );
	} |
	Factor {
		$$ = new FactorWithNeg( InputLoc(@1), $1 );
	};

/* The highest level in precedence. Atomic machines such as references to other
 * machines, literal machines, regular expressions or Expressions in side of
 * parenthesis. */
Factor:
	TK_Literal {
		// Create a new factor node going to a concat literal. */
		$$ = new Factor( new Literal( InputLoc(@1), $1, Literal::LitString ) );
	} |
	AlphabetNum {
		// Create a new factor node going to a literal number. */
		$$ = new Factor( new Literal( InputLoc(@1), $1, Literal::Number ) );
	} |
	TK_Word {
		// Find the named graph.
		GraphDictEl *gdNode = pd->graphDict.find( $1 );
		if ( gdNode == 0 ) {
			// Recover by returning null as the factor node.
			error(@1) << "graph lookup of \"" << $1 << "\" failed" << endl;
			$$ = 0;
		}
		else if ( gdNode->isInstance ) {
			// Recover by retuning null as the factor node.
			error(@1) << "references to graph instantiations not allowed "
					"in expressions" << endl;
			$$ = 0;
		}
		else {
			// Create a factor node that is a lookup of an expression.
			$$ = new Factor( gdNode->value );
		}
	} |
	RE_SqOpen RegularExprOrData RE_SqClose {
		// Create a new factor node going to an OR expression. */
		$$ = new Factor( new ReItem( InputLoc(@1), $2, ReItem::OrBlock ) );
	} |
	RE_SqOpenNeg RegularExprOrData RE_SqClose {
		// Create a new factor node going to a negated OR expression. */
		$$ = new Factor( new ReItem( InputLoc(@1), $2, ReItem::NegOrBlock ) );
	} |
	RE_Slash RegularExpr RE_Slash {
		// Create a new factor node going to a regular exp.
		$$ = new Factor( $2 );
	} |
	RangeLit TK_DoubleDot RangeLit {
		// Create a new factor node going to a range. */
		$$ = new Factor( new Range( $1, $3 ) );
	} |
	'(' Join ')' {
		$$ = new Factor( $2 );
	};

/* Garble up to the closing brace of a parenthesized expression. */
Factor: '(' error ')' { $$ = 0; yyerrok; };

/* Any form of a number that can be used as a basic machine. */
AlphabetNum:
	TK_UInt | TK_PInt | TK_NInt | TK_Hex ;

/* Collect a block of C/C++ code in curly braces into a single string. 
 * Does not interpret special keywords. */
CurlyBlock:
	'{' CurlyBlockList '}' {
		codeBlockBuf.append( 0 );
		$$ = new char[codeBlockBuf.length];
		memcpy( $$, codeBlockBuf.data, codeBlockBuf.length );
	};

/* List of curly block items, appens to the codeBlockList. */
CurlyBlockList:
	CurlyBlockList CurlyBlockAny {
		codeBlockBuf.append( $2, strlen($2) );
		delete[] $2;
	} |
	CurlyBlockList CurlyBlockSpecial {
		codeBlockBuf.append( $2, strlen($2) );
	} |
	/* Nothing */ {
		codeBlockBuf.clear();
	};

/* Anything in a curly block that is dynamically allocated by the scanner. */
CurlyBlockAny:
	IL_WhiteSpace | IL_Comment | IL_Literal | IL_Symbol |
	TK_UInt | TK_PInt | TK_NInt | TK_Hex | TK_Word;

/* Anything in a curly block that is not dynamically allocated. */
CurlyBlockSpecial:
	'[' { $$ = "["; } | 
	']' { $$ = "]"; } | 
	',' { $$ = ","; } | 
	';' { $$ = ";"; } |
	KW_Stack { $$ = "fstack"; } |
	KW_Hold { $$ = "fhold"; } |
	KW_Char { $$ = "fc"; } |
	KW_Goto { $$ = "fgoto"; } |
	KW_Call { $$ = "fcall"; } |
	KW_Ret { $$ = "fret"; };


/* Collect a block of code in curly braces into a specially formatted string
 * that contains null termined sequencees of text and commands. */
CodeBlock:
	'{' CodeBlockList '}' {
		codeBlockBuf.append( CB_END );
		$$ = new char[codeBlockBuf.length];
		memcpy( $$, codeBlockBuf.data, codeBlockBuf.length );
	};

/* List of items in a code block. */
CodeBlockList:
	CodeBlockList CodeBlockItem |
	/* Nothing */ {
		codeBlockBuf.clear();
	};

/* Items in a code block. */
CodeBlockItem:
	CodeBlockAny {
		/* Add a text segment, delete the data. */
		codeBlockBuf.append( CB_TEXT );
		codeBlockBuf.append( $1, strlen($1) );
		codeBlockBuf.append( 0 );
		delete[] $1;
	} |
	CodeBlockSpecial {
		/* Add a text segment. */
		codeBlockBuf.append( CB_TEXT );
		codeBlockBuf.append( $1, strlen($1) );
		codeBlockBuf.append( 0 );
	} |
	CodeBlockInterpret;

/* Uninteresting tokens in a code block. Data allocated by scanner. */
CodeBlockAny:
	IL_WhiteSpace | IL_Comment | IL_Literal | IL_Symbol |
	TK_UInt | TK_PInt | TK_NInt | TK_Hex | TK_Word;

/* Symbols in a code block, no data allocated. */
CodeBlockSpecial:
	KW_Stack { $$ = "fstack"; } |
	'[' { $$ = "["; } | 
	']' { $$ = "]"; } | 
	',' { $$ = ","; } | 
	';' { $$ = ";"; };

/* Interpreted statements in a code block. */
CodeBlockInterpret:
	KW_Hold ';' {
		codeBlockBuf.append( CB_HOLD );
		codeBlockBuf.append( 0 );
	} |
	KW_Char {
		codeBlockBuf.append( CB_CHAR );
		codeBlockBuf.append( 0 );
	} |
	KW_Goto IL_WhiteSpace TK_Word ';' { 
		codeBlockBuf.append( CB_JUMP );
		codeBlockBuf.append( $3, strlen($3) );
		codeBlockBuf.append( 0 );
		delete[] $2;
		delete[] $3;
	} | 
	KW_Call IL_WhiteSpace TK_Word ';' {
		codeBlockBuf.append( CB_CALL );
		codeBlockBuf.append( $3, strlen($3) );
		codeBlockBuf.append( 0 );
		delete[] $3;
	} | 
	KW_Ret ';' {
		codeBlockBuf.append( CB_RET );
		codeBlockBuf.append( 0 );
	};


/* Collect a block of structure data in curly braces into a specially formatted
 * string that contains null termined sequencees of text and specials. */
StructBlock:
	'{' StructBlockList '}' {
		codeBlockBuf.append( CB_END );
		$$ = new char[codeBlockBuf.length];
		memcpy( $$, codeBlockBuf.data, codeBlockBuf.length );
	};

/* List of items in a struct block. */
StructBlockList:
	StructBlockList StructBlockItem |
	/* Nothing */ {
		codeBlockBuf.clear();
	};

/* Items in a struct block. */
StructBlockItem:
	StructBlockAny {
		/* Add a text segment, delete the data. */
		codeBlockBuf.append( CB_TEXT );
		codeBlockBuf.append( $1, strlen($1) );
		codeBlockBuf.append( 0 );
		delete[] $1;
	} |
	StructBlockSymbol {
		/* Add a text segment. */
		codeBlockBuf.append( CB_TEXT );
		codeBlockBuf.append( $1, strlen($1) );
		codeBlockBuf.append( 0 );
	} |
	StructBlockInterpret;

/* Uninteresting tokens in a struct block. Data allocated by scanner. */
StructBlockAny:
	IL_WhiteSpace | IL_Comment | IL_Literal | IL_Symbol |
	TK_UInt | TK_PInt | TK_NInt | TK_Hex | TK_Word;

/* Symbols in a struct block, no data allocated. */
StructBlockSymbol:
	'[' { $$ = "["; } | 
	']' { $$ = "]"; } | 
	',' { $$ = ","; } | 
	';' { $$ = ";"; };

/* Interpreted statements in a struct block. */
StructBlockInterpret:
	KW_Stack '[' TK_UInt ']' ';' {
		codeBlockBuf.append( CB_STACK );
		codeBlockBuf.append( $3, strlen($3) );
		codeBlockBuf.append( 0 );
	};


/* Parser for regular expression fsms. Any number of expression items which
 * generally gives a machine one character long or one character long stared. */
RegularExpr:
	RegularExpr RegularExprItem {
		// An optimization to lessen the tree size. If a non-starred char is directly
		// under the left side on the right and the right side is another non-starred
		// char then paste them together and return the left side. Otherwise
		// just put the two under a new reg exp node.
		if ( $2->type == ReItem::Data && !$2->star &&
			$1->type == RegExpr::RecurseItem &&
			$1->item->type == ReItem::Data && !$1->item->star )
		{
			// Append the right side to the right side of the left and toss 
			// the right side.
			$1->item->str += $2->str;
			delete $2;
			$$ = $1;
		}
		else {
			$$ = new RegExpr( $1, $2 );
		}
	} |
	/* Nothing */ {
		// Can't optimize the tree.
		$$ = new RegExpr();
	};

/* RegularExprItems can be a character spec with an optional staring of the char. */
RegularExprItem:
	RegularExprChar RE_Star {
		$1->star = true;
		$$ = $1;
	} |
	RegularExprChar {
		$$ = $1;
	};

/* A character spec can be a set of characters inside of square parenthesis,
 * a dot specifying any character or some explicitly stated character. */
RegularExprChar:
	RE_SqOpen RegularExprOrData RE_SqClose {
		$$ = new ReItem( InputLoc(@1), $2, ReItem::OrBlock );
	} |
	RE_SqOpenNeg RegularExprOrData RE_SqClose {
		$$ = new ReItem( InputLoc(@1), $2, ReItem::NegOrBlock );
	} |
	RE_Dot {
		$$ = new ReItem( InputLoc(@1), ReItem::Dot );
	} |
	RE_Char {
		$$ = new ReItem( InputLoc(@1), $1[0] );
	};

/* The data inside of a [] expression in a regular expression. Accepts any
 * number of characters or ranges. */
RegularExprOrData:
	RegularExprOrData RegularExprOrChar {
		// An optimization to lessen the tree size. If an or char is directly
		// under the left side on the right and the right side is another or
		// char then paste them together and return the left side. Otherwise
		// just put the two under a new or data node.
		if ( $2->type == ReOrItem::Data &&
				$1->type == ReOrBlock::RecurseItem &&
				$1->item->type == ReOrItem::Data )
		{
			// Append the right side to right side of the left and toss
			// the right side.
			$1->item->str += $2->str;
			delete $2;
			$$ = $1;
		}
		else {
			// Can't optimize, put the left and right under a new node.
			$$ = new ReOrBlock( $1, $2 );
		}
	} | 
	/* Nothing */ {
		$$ = new ReOrBlock();
	};


/* A single character inside of an or expression. Can either be a character
 * or a set of characters. */
RegularExprOrChar:
	RE_Char {
		$$ = new ReOrItem( $1[0] );
	} |
	RE_Char RE_Dash RE_Char {
		$$ = new ReOrItem( $1[0], $3[0] );
	};

RangeLit:
	TK_Literal {
		// Range literas must have only one char.
		if ( strlen($1) != 1 ) {
			// Recover by using the literal anyways.
			error(@1) << "literal used in range must be of length 1" << endl;
		}
		$$ = new Literal( InputLoc(@1), $1, Literal::LitString );
	} |
	AlphabetNum {
		// Create a new literal number.
		$$ = new Literal( InputLoc(@1), $1, Literal::Number );
	};

%%

/* Try to do a definition, common to assignment and instantiation. Warns about 
 * instances other than main not being implemented yet. */
void tryMachineDef( const BISON_YYLTYPE &loc, char *name, Join *join, bool isInstance )
{
	GraphDictEl *newEl = pd->graphDict.insert( name );
	if ( newEl != 0 ) {
		/* New element in the dict, all good. */
		newEl->value = new VarDef( name, join );
		newEl->isInstance = isInstance;
		newEl->line = loc.first_line;
		newEl->col = loc.first_column;

		/* It it is an instance, put on the instance list. */
		if ( isInstance )
			pd->instanceList.append( newEl );
	}
	else {
		// Recover by ignoring the duplicate.
		error(loc) << "fsm \"" << name << "\" previously defined" << endl;
	}
	pd->machineGiven = true;
}

void yyerror( char *err )
{
	error(yylloc.first_line, yylloc.first_column) << err << endl;
}
